import os
import logging
import re
from uuid import uuid1
from random import randrange

from croniter import croniter
from datetime import datetime, date, timedelta

from p2.agent.client import Client as AgentClient
from p2.agent.cronjob import Cronjob
from p2.store.client import Client as StoreClient
from p2.scheduler.Config import Config
from p2.scheduler.Test import P2test
from p2.scheduler.Shifts import Shifts
from p2.driver.k8s.k8s import get_current_context


class Scheduler:
    """A class to schedule testing campaigns.
    The workflow is roughly divided into these steps:
    - first we get a list of what tests (experiments on targets) we would like to run
    - then we try to find a free slot to run them
    - last we schedule the execution of these tests as Kubernetes cronjobs in the same
      context the scheduler is running in.

    A scheduler object accepts as attributes:
    - a logger object: (not mandatory), the default logging lib is used if not passed
    - a tag: (not mandatory) used to disambiguate between multiple scheduler instances.
    """

    def __init__(self, logger=None, tag=None):

        try:
            self.log = logger.getLogger() if logger else logging.getLogger(__name__)

            store_url = os.environ.get('P2_STORE_URL')
            agent_url = os.environ.get('P2_AGENT_URL')
            self.log.debug(f"Store URL {store_url}")
            self.log.debug(f"Agent URL {agent_url}")

            self.p2_store_client = StoreClient(store_url)
            if self.p2_store_client is None:
                raise Exception("Error getting a p2.store client object")

            self.p2_agent_client = AgentClient(agent_url)
            if self.p2_agent_client is None:
                raise Exception("Error getting a p2.agent client object")

            self.config = Config(tag = tag)
            self.shifts = Shifts(tag = tag)
            self.to_be_scheduled = list()
            self.context = get_current_context().get('name')

        except Exception as err:
            self.log.error(err)
            return None

    def schedule_campaign(self, campaign_id: int):
        """
        - retrieves the campaign with id
        - get targets
        - get experiments
        - see if experiments have to be scheduled...
        """
        self.campaign_id = campaign_id
        self.campaign    = self._get_campaign()
        # TODO
        # the filter on the campaign has to be updated...
        # create a new campaign, better: add this feature to the p2.ui.campaigns app...
        self.targets     = self.p2_store_client.select(
                               table='resources',
                               filter=self.campaign['resources']
                           )
        self.experiments = self.p2_store_client.select(
                               table='experiments',
                               filter=self.campaign['experiments']
                           )

        """
        delete old sessions, taking into consideration
        the next time he is going to run...
        actually we delete just the cronjob, keep the
        session record...
        """
        old_sessions = self.p2_store_client.select(
            table='exp_session',
            filter={'start_time|< now() - interval ': '357 days'},
            fetch=['exp_id', 'start_time']
        )
        for session in old_sessions:
            try:
                # self.p2_store_client.delete(table='exp_session', key='exp_id',
                                            # value=session['exp_id'])
                self.p2_agent_client.delete_cronjob(context=self.context, id=session['exp_id'])
            except Exception as e:
                self.log.error(f"ERROR deleting old sessions: {e}")

        self._get_schedule()
        self._schedule_all()
        return 1


    def _get_schedule(self):
        for target in self.targets:
            for experiment in self.experiments:
                if  experiment['type'] == 'experiment' and \
                    'profile' in target and \
                    'target' in experiment['profile'] and \
                    target['profile']['name'] == experiment['profile']['target']['name']:
                    self._add_to_schedule(target=target, experiment=experiment)
                elif experiment['type'] == 'template':
                    # if it's a template then any target is valid...
                    self._add_to_schedule( target=target, experiment=experiment)


    def _add_to_schedule(self, target: dict, experiment: dict):
        self.log.info("See if experiment {} on target {} has to be scheduled...".format(
            experiment['filename'], target['name']))

        last_session = self._get_last_session_for_target(experiment=experiment,
                                                         target=target)
        if last_session is not None and 'status' in last_session:
            min_interval = self.min_interval(target, experiment, last_session['status'])
            expired_by = date.today() - date.fromordinal(last_session['scheduled_day'])
            if 'end_time' in last_session and last_session['end_time'] != None:
                expired_by = date.today() - datetime.fromisoformat(last_session['end_time']).date()

            if expired_by.days > min_interval:
                self.to_be_scheduled.append( P2test(
                        campaign = self.campaign,
                        target = target,
                        experiment = experiment,
                        expired = expired_by.days - min_interval))
                return True
        else:
            # no previous run for this experiment on this target, initializing
            # expiration time...
            self.to_be_scheduled.append( P2test(
                campaign = self.campaign,
                target = target,
                experiment = experiment,
                # expired = self.config.target_priority_period[target_priority]
                expired = 100
                ))
            return True
        return False


    def _schedule_all(self):
        """schedule tests sorted by expiretion time"""
        not_scheduled=list()
        to_be_scheduled_sorted = sorted(self.to_be_scheduled,
                                        key=lambda item: item.expired,
                                        reverse=True)
        for test in to_be_scheduled_sorted:
            duration = test.experiment['duration'] if 'duration' in test.experiment else 30
            if (start_at := self.shifts.get_session_shift(duration)) is not None:
                day = date(
                    day=start_at['day'],
                    month=start_at['month'],
                    year=start_at['year']
                )
                self._schedule_this(test, day, start_at)
            else:
                self.log.warning("No days left to schedule experiment {} on target {}".format(
                    test.experiment['filename'], test.target['name']))
                not_scheduled.append(test)
        return not_scheduled


    def _schedule_this(self, test: P2test, day: date, start_at: dict):
        """Finds the best slot in the day for this test

            Gets the max(next_run_time) of the day
            adds that test expected duration and round up to next hour
            check hour is in test shift
            in case of a template:
            - template profile should hold the template static parameters
            in case these are set in the test campaign, they override the ones
            set in the template. This is handled by the agent (cron) job.
            - session profile should hold the dynamic params
            - Current dynamic params supported:
                - target: target profile
            """
        try:
            mins = "%02d" % start_at['start_minute']
            hour = "%02d" % start_at['start_hour']
            dom  = "%02d" % start_at['day']
            month= "%02d" % start_at['month']
            schedule = f"{mins} {hour} {dom} {month} *"
            cron = croniter(schedule, datetime.now())
            next_run_time = cron.get_next(datetime)
            exp_id = test.target['name'] + '-' + str(uuid1())[:8]
            session = {
                    'filepath': test.experiment['filename'],
                    'campaign_id': self.campaign_id,
                    'context': self.context,
                    'schedule': schedule,
                    'next_run_time': str(next_run_time),
                    'mins': mins,
                    'hour': hour,
                    'dom': dom,
                    'month': month,
                    'dow': '*',
                    'year': start_at['year'],
                    'profile':{'target':test.target}, # TODO : this is going to be deleted soon...
                    'resource_id': test.target['id'],
                    'status': 'loaded'
            }
            exp_id = self.p2_store_client.create_session(exp_id, session)
            if exp_id != None:
                cronjob = Cronjob(
                        id = exp_id,
                        crontab = schedule,
                        filename = test.experiment['filename'],
                        driver = test.experiment['type'],
                        campaign_id = self.campaign_id
                        )
                ret = self.p2_agent_client.create_cronjob(self.context, cronjob)
                if ret != exp_id:
                    raise Exception("Couldn't add cronjob")

                self.p2_store_client.update_session(exp_id, {'status':'scheduled'})

                if test in self.to_be_scheduled:
                    self.to_be_scheduled.remove(test)

                self.log.info("Session {} scheduled".format(exp_id))
            else:
                self.log.error(f"Error scheduling session for experiment {session['filepath']}")
            return exp_id
        except Exception as e:
            self.log.error(f"Couldn't schedule test, exception: {e}")
            return None


    def min_interval(self, target, experiment, status):
        """computes min interval (in days ) between tests
            of same type on same target """

        exp_profile = experiment.get('profile',{'weight':1.0})
        target_priority = self._get_target_priority(target)
        weight = exp_profile.get('weight', 1.0)
        if status == 'failed':
            if weight > 0.9: weight = weight - 0.5
        return int(self.config.target_priority_period[target_priority] * weight)


    def _get_target_priority(self, target):
        target_priority = self.config.targets_priority.get(target['name'],
                            len(self.config.target_priority_period))
        return target_priority-1


    def _get_campaign(self):
        campaign = self.p2_store_client.get_campaign_of_id(self.campaign_id)
        if campaign is None:
            raise Exception(f"No campaign with id {id} found")

        if campaign['resources'] is None:
            raise Exception(f"Error, no resources set for campaign {id}")

        if campaign['experiments'] is None:
            raise Exception(f"Error, no experiments set for campaign {id}")
        return campaign


    def _get_last_session_for_target(self, experiment: dict, target: dict):
        """Returns the last session for the passed target and experiment, no matter the status """

        sessions = self.p2_store_client.select(
            table='exp_session',
            filter={
                'filepath': experiment['filename'],
                'resource_id': target['id']
            })
        if sessions is None or len(sessions) == 0:
            return None

        for s in sessions:
            s['scheduled_day'] = date(s['year'], int(s['month']),
                                      int(s['dom'])).toordinal()

        sessions_sorted = sorted(sessions, key=lambda session:
                                 session['scheduled_day'],
                                 reverse=True)
        return sessions_sorted[0]
